
""" KnackHQ REST Client. """


import collections
import json
import os

import certifi
import urllib3
from .knackhq import KnackHQObject


class KnackHQClient(collections.Iterable):
    """ KnackHQ root client.

        Use arguments or ENV variables to initialize client.

        Optional ENV variables are:
            * KNACKHQ_APP_ID
            * KNACKHQ_API_KEY
            * KNACKHQ_ENDPOINT

        Arguments:
            app_id   (str):  Application ID string
            api_key  (str):  API key
            endpoint (str):  KnackHQ endpoint (default: https://api.knackhq.com/v1)
    """
    @property
    def _headers(self):
        """ Helper to define request headers. """
        return {
            "Content-Type": "application/json",
            "X-Knack-Application-Id": self._app_id,
            "X-Knack-REST-API-Key": self._api_key
        }

    def __init__(self, app_id=None, api_key=None, endpoint=None):
        self._app_id = app_id or os.getenv('KNACKHQ_APP_ID')
        self._api_key = api_key or os.getenv('KNACKHQ_API_KEY')
        self._endpoint = endpoint or os.getenv('') \
            or os.getenv('KNACKHQ_ENDPOINT') \
            or 'https://api.knackhq.com/v1'

    def __repr__(self):
        return "<KnackHQClient %s>" % self._endpoint

    def __iter__(self):
        return self.get_objects()

    def __len__(self):
        len(list(self))

    def endpoint(self, *path):
        """ Helper to get URL of object or its children

            Arguments:
                path (tuple):  Optional additional children

            Returns:
                URL string for object-resource.
        """
        return os.path.join(self._endpoint, *path)

    def request(self, url, verb='GET', **kwargs):
        """ Send a request to the API.

            Arguments:
                url    (str):   URL for request
                verb   (str):   Optional verb of request
                kwagrs (dict):  Optional additional request arguments

            Returns:
                JSON response of request.

            Raises:
                ValueError on bad response.
        """
        http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where())
        resp = http.request(verb, url, headers=self._headers, **kwargs)
        try:
            return json.loads(resp.data)
        except ValueError as err:
            err.message += "\n%s" % resp.data
            raise err

    def get_objects(self, name=None):
        """ Yield a collection of objects. If a name is supplied only objects
            with that name are returned.

            Arguments:
                name (str):  Optional name filter for objects

            Returns:
                Iterator of KnackHQObjects
        """
        for item in self.request(self.endpoint('objects')).get('objects', []):
            if name is None or item['name'] == name:
                yield KnackHQObject(self, item['key'])

    def get_object(self, key=None, name=None):
        """ Get object by key or name. If a key is not provided it
            will be found by filtering on all objects. It is more
            efficient to use the key for getting objects.

            Arguments:
                key  (str):  Object key
                name (str):  Object name

            Returns:
                KnackHQObject instance.

            Raises:
                KeyError if object not found when finding by name.
        """
        if key is None:
            key = self.get_object_key(name)
        return KnackHQObject(self, key)

    def get_object_key(self, name):
        """ Helper to get object key from object name.

            Arguments:
                name (str):  Name of object

            Returns:
                Key of object.

            Raises:
                KeyError on object not found.
        """
        objects = list(self.get_objects(name))
        if len(objects) == 1:
            return objects[0].key
        elif len(objects) > 1:
            raise KeyError("More than one object named '%s'" % name)
        raise KeyError



""" KnackHQ Objects/Records """


import collections
import os
import json
import urllib

from urlparse import urlparse


class KnackHQRecord(collections.Mapping):
    """ KnackHQ Record.
    """
    @property
    def record(self):
        """ Memoized response of a GET on the record. """
        if self._record is None:
            self._record = self.object.client.request(self.url.geturl())
        return self._record

    @property
    def object(self):
        """ Get parent KnackHQObject. """
        return self._object

    def __init__(self, knackobj, url, record=None):
        self._object = knackobj
        self._record = record
        self.url = urlparse(url)

    def __getitem__(self, key):
        return self.record[key]

    def __iter__(self):
        return iter(self.record)

    def __len__(self):
        return len(self.record)

    def get_field(self, key):
        """ Get value of record field by field key. """
        fields = [x for x in self.object['fields'] if x['name'] == key]
        if len(fields) == 1:
            field_key = fields[0]['key']
            field_raw = "%s_raw" % field_key
            is_field = lambda x: x in (field_key, field_raw)
            return dict([(key, val) for key, val in self.record.iteritems() if is_field(key)])

        raise KeyError("More than one field named '%s'" % key)


class KnackHQObject(collections.Iterable):
    """ KnackHQ Object (ie, table).

        Arguments:
            client (Client):  KnackHQ Client instance
            url    (str):     URL to object resource

        Iterate over records in an object with:

            for record in object:
                # do something

        Iterate over filtered records in an object with:

            filters = [
                {
                    field: 'field_1',
                    operator: 'is',
                    value: 'test'
                }, {
                    field: 'field_2',
                    operator: 'is not blank'
                }
            ]

            for record in object.where(*filters):
                # do something
    """
    @property
    def object(self):
        """ Memoized response of a GET on the object. """
        if self._object is None:
            self._object = self.client.request(self.url.geturl())['object']
        return self._object

    def __init__(self, client, key):
        self.client = client
        self._len = None
        self._object = None
        self.key = key
        self.url = urlparse(self.client.endpoint('objects', key))

    def __repr__(self):
        return "<%s %s>" % (self['name'], self.url.path)

    def __getitem__(self, key):
        return self.object[key]

    def __iter__(self):
        return self.get_records()

    def __len__(self):
        if self._len is None:
            page = self.get_records().next()
            self._len = page['total_records']
        return self._len

    def get_records(self, **kwargs):
        """ Get records from a REST request.

            Arguments:
                page          (int):   Optional page number
                rows_per_page (int):   Optional rows per page number
                sort_field    (str):   Optional sort field key
                sort_order    (str):   Optional sort order
                filters       (list):  Optional list of filter dicts

            Returns:
                REST response JSON dict.
        """
        url = self.client.endpoint('objects', self.key, 'records')
        url += '?'
        for key, val in kwargs.iteritems():
            if key == 'filters':
                val = urllib.quote_plus(json.dumps(val))
            url += "%s=%s&" % (key, val)
        try:
            page = self.client.request(url)
        except ValueError:
            raise StopIteration
        record_url = urlparse(url).path
        for record in page['records']:
            record_url = os.path.join(self.url.geturl(), 'records', record['id'])
            yield KnackHQRecord(self, record_url, record)
        kwargs['page'] = int(page['current_page']) + 1
        self.get_records(**kwargs)

    def get_record(self, record_id):
        """ Get record by ID.

            Arguments:
                record_id (str):  Record ID string

            Returns:
                REST response JSON dict.
        """
        url = self.client.endpoint(self.url.path.lstrip('/'), 'records', record_id)
        return self.client.request(url)

    def keys(self):
        """ Return keys of KnackHQ object definiton. """
        return self.object.keys()




import os
import re
from setuptools import setup

NAME    = "knackhq"
AUTHOR  = "amancevice"
EMAIL   = "smallweirdnum@gmail.com"
DESC    = "Interact with KnackHQ API"

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2.7",
    "Topic :: Utilities" ]
REQUIRES = [
    "certifi>=2016.2.28",
    "urllib3>=1.12" ]
TEST_REQUIRES = ["nose", "mock"]

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def version():
    search = r"^__version__ *= *['\"]([0-9.]+)['\"]"
    initpy = read("./%s/__init__.py" % NAME)
    return re.search(search, initpy, re.MULTILINE).group(1)

setup(
    name                 = NAME,
    version              = version(),
    author               = AUTHOR,
    author_email         = EMAIL,
    packages             = [ NAME ],
    package_data         = { "%s" % NAME : ['README.md'] },
    include_package_data = True,
    url                  = 'http://www.smallweirdnumber.com',
    description          = DESC,
    long_description     = read('README.md'),
    classifiers          = CLASSIFIERS,
    install_requires     = REQUIRES,
    test_requires        = TEST_REQUIRES,
    test_suite           = "nose.collector" )
